from collections import namedtuple
from itertools import chain

from rpy2.robjects.packages import importr
from rpy2.robjects import numpy2ri, pandas2ri
from rpy2.robjects.conversion import localconverter

import rpy2.robjects as ro

base = importr('base')
forecast = importr('forecast')


def get_r_model(ts, period1, period2, h, coeffs, armethod):

    with localconverter(ro.default_converter + numpy2ri.converter):
        r_ts = ro.conversion.py2rpy(ts)

    r_msts = forecast.msts(r_ts, **{
        'seasonal.periods': ro.IntVector([period1, period2]),
    })

    fit = forecast.dshw(r_msts,
                        armethod=armethod,
                        h=h, **coeffs)

    return convert_model_props(fit)


def convert_model_props(fit):
    fitd = dict(zip(fit.names, list(fit)))

    pnames = ['alpha', 'beta', 'gamma', 'omega', 'phi']
    snames = ['mean', 'fitted', 'residuals']

    with localconverter(ro.default_converter + numpy2ri.converter + pandas2ri.converter):
        model = namedtuple('Params', pnames + snames)(
            *chain(
                (float(fitd['model'].rx[pname][0]) for pname in pnames),
                (ro.conversion.rpy2py(fitd[sname]) for sname in snames),
            )
        )

    return model



import numpy as np

import torch


def seasonality_ts(n, s_len=24):

    freq = n / s_len

    t = np.arange(n) / n
    c1 = 1.0 * np.sin(2 * np.pi * t * freq)
    c2 = .4 * np.sin(2 * np.pi * 15 * t)

    noise = np.random.rand(n)

    return c1 + c2 + noise


def trend(n, steepness=1.2):
    return np.arange(n) / (n**steepness) + np.random.rand(n) * 0.1


class DataSetOneFixture(object):

    n = 24 * 7 * 4 * 4
    h = 24 * 7 * 4
    level = 50
    period1 = 24
    period2 = 24 * 7

    @property
    def ts(self):
        np.random.seed(0)
        return (self.level +
                trend(self.n, steepness=0.9) +
                seasonality_ts(self.n, self.period1) +
                seasonality_ts(self.n, self.period2))

    @property
    def torch_ts(self):
        return torch.from_numpy(self.ts).float()


class DataSet2FeatureFixture(object):

    n = 24 * 7 * 4 * 4
    h = 24 * 7 * 4
    level1 = 50
    level2 = 150
    period1 = 24
    period2 = 24 * 7

    @property
    def ts(self):
        np.random.seed(1)
        a = (self.level1 +
             trend(self.n, steepness=0.9) +
             seasonality_ts(self.n, self.period1) +
             seasonality_ts(self.n, self.period2)).reshape(-1, 1)
        np.random.seed(0)
        b = (self.level2 +
             trend(self.n, steepness=1.0) +
             seasonality_ts(self.n, self.period1) +
             seasonality_ts(self.n, self.period2)).reshape(-1, 1)
        return np.concatenate([a, b], axis=1)

    @property
    def torch_ts(self):
        return torch.from_numpy(self.ts).float()


class DataSet2BatchFixture(object):

    n = 24 * 7 * 4 * 4
    h = 24 * 7 * 4
    level1 = 50
    level2 = 150
    period1 = 24
    period2 = 24 * 7

    @property
    def ts(self):
        np.random.seed(1)
        a = (self.level1 +
             trend(self.n, steepness=0.9) +
             seasonality_ts(self.n, self.period1) +
             seasonality_ts(self.n, self.period2)).reshape(1, -1)
        np.random.seed(0)
        b = (self.level2 +
             trend(self.n, steepness=1.0) +
             seasonality_ts(self.n, self.period1) +
             seasonality_ts(self.n, self.period2)).reshape(1, -1)
        return np.concatenate([a, b], axis=0)

    @property
    def torch_ts(self):
        return torch.from_numpy(self.ts).float()


class DataSet2Batch2FeatureFixture(object):

    n = 24 * 7 * 4 * 4 * 2
    h = 24 * 7 * 4
    level1 = 50
    level2 = 150
    period1 = 24
    period2 = 24 * 7

    @property
    def ts(self):
        np.random.seed(1)
        a = (self.level1 +
             trend(self.n, steepness=0.9) +
             seasonality_ts(self.n, self.period1) +
             seasonality_ts(self.n, self.period2)).reshape(2, -1, 1)
        np.random.seed(0)
        b = (self.level2 +
             trend(self.n, steepness=1.0) +
             seasonality_ts(self.n, self.period1) +
             seasonality_ts(self.n, self.period2)).reshape(2, -1, 1)
        return np.concatenate([a, b], axis=2)

    @property
    def torch_ts(self):
        return torch.from_numpy(self.ts).float()



import torch
import torch.nn as nn

from torches.utils.modelparams import default_param, pytorch_diff
from torches.utils.seasonality import add_seasindex, mult_seasindex


DEFAULT_PARAM_VALUE = 1e-03


make_obj = lambda **kwargs: type('Res', (object,), kwargs)()


class DSHWModule(nn.Module):
    """
    Double Sesonal Holt Winters.

    This module provides double seasonal exponential smoothing for 3d data: <batch, series_len, series_index>.
    In other words, you can learn HW-coefficients for multiple timeseries at once with batching support.
    Losses for each series can be arbitrary.

    Reference implementation can be found at https://github.com/robjhyndman/forecast/blob/v8.7/R/dshw.r.

    WARN: max_h is used to compute forecast horizons. You have to cut it to :h for loss computation
    """
    # TODO: it probably doesn't make sense to have multiple "hs", since we can just predict for the biggest
    # and then cut the predictions in the loss function
    def __init__(self, input_dim, period1, period2, hs,
                 alphas=None, betas=None, gammas=None, omegas=None, phis=None,
                 enable_ar=False, is_multiplicative=True):
        super().__init__()
        assert len(hs) == input_dim

        # TODO: alphas=False => requires_grad=False

        self.input_dim = input_dim
        self.period1 = period1
        self.period2 = period2
        self.hs = torch.tensor(hs, requires_grad=False)
        self.enable_ar = enable_ar
        self.is_multiplicative = is_multiplicative

        alphas = default_param(input_dim, alphas, DEFAULT_PARAM_VALUE)
        betas = default_param(input_dim, betas, DEFAULT_PARAM_VALUE)
        gammas = default_param(input_dim, gammas, DEFAULT_PARAM_VALUE)
        omegas = default_param(input_dim, omegas, DEFAULT_PARAM_VALUE)
        phis = default_param(input_dim, phis, DEFAULT_PARAM_VALUE)

        self.alphas = torch.nn.Parameter(alphas)
        self.betas = torch.nn.Parameter(betas)
        self.gammas = torch.nn.Parameter(gammas)
        self.omegas = torch.nn.Parameter(omegas)
        self.phis = torch.nn.Parameter(phis)

    def mult_init_params(self, y):
        assert y.ndimension() == 3, "3-dim input is expected"

        bs, n, f = y.size()

        # Starting values
        I1 = mult_seasindex(y, self.period1)
        w1 = mult_seasindex(y, self.period2)

        ratio = self.period2 // self.period1
        w1 = w1 / I1.repeat(1, ratio, 1)

        # TODO: torch.pad
        x = torch.cat([torch.zeros((bs, 1, f)), pytorch_diff(y[:, :self.period2, :])], dim=1)
        t = torch.mean(((y[:, :self.period2, :] - y[:, self.period2:(2 * self.period2), :]) / self.period2) + x, dim=1) / 2  # trend
        s = torch.mean(y[:, :(2 * self.period2), :], dim=1) - (self.period2 + 0.5) * t  # level

        return I1, w1, t, s

    def add_init_params(self, y):
        assert y.ndimension() == 3, "3-dim input is expected"

        bs, n, f = y.size()

        # Starting values
        I1 = add_seasindex(y, self.period1)
        w1 = add_seasindex(y, self.period2)

        ratio = self.period2 // self.period1
        w1 = w1 - I1.repeat(1, ratio, 1)

        x = torch.cat([torch.zeros((bs, 1, f)), pytorch_diff(y[:, :self.period2, :])], dim=1)
        t = torch.mean(((y[:, :self.period2, :] - y[:, self.period2:(2 * self.period2), :]) / self.period2) + x, dim=1) / 2  # trend
        s = torch.mean(y[:, :(2 * self.period2), :], dim=1) - (self.period2 + 0.5) * t  # level

        return I1, w1, t, s

    def get_coeffs(self):
        return {c: torch.sigmoid(getattr(self, c)) for c in ['alphas', 'betas', 'gammas', 'omegas', 'phis']}

    def add_update(self, y, Ic, wc, s, t):

        bs, n, f = y.size()

        alphas = torch.sigmoid(self.alphas)
        betas = torch.sigmoid(self.betas)
        gammas = torch.sigmoid(self.gammas)
        omegas = torch.sigmoid(self.omegas)

        yhat = torch.empty((bs, n, f))

        for i in range(n):
            yhat[:, i, :] = (s + t) + Ic[:, i % self.period1, :] + wc[:, i % self.period2, :]
            snew = alphas * (y[:, i, :] - (Ic[:, i % self.period1, :] + wc[:, i % self.period2, :])) + (1 - alphas) * (s + t)
            tnew = betas * (snew - s) + (1 - betas) * t

            Ico = Ic.clone()
            wco = wc.clone()

            Ico[:, i % self.period1, :] = gammas * (y[:, i, :] - (snew - wc[:, i % self.period2, :])) + (1 - gammas) * Ic[:, i % self.period1, :]
            wco[:, i % self.period2, :] = omegas * (y[:, i, :] - (snew - Ic[:, i % self.period1, :])) + (1 - omegas) * wc[:, i % self.period2, :]
            Ic = Ico
            wc = wco

            s = snew
            t = tnew

        return yhat, Ic, wc, s, t

    def mult_update(self, y, Ic, wc, s, t):

        bs, n, f = y.size()

        alphas = torch.sigmoid(self.alphas)
        betas = torch.sigmoid(self.betas)
        gammas = torch.sigmoid(self.gammas)
        omegas = torch.sigmoid(self.omegas)

        yhat = torch.empty((bs, n, f))

        for i in range(n):
            yhat[:, i, :] = (s + t) * Ic[:, i % self.period1, :] * wc[:, i % self.period2, :]
            snew = alphas * (y[:, i, :] / (Ic[:, i % self.period1, :] * wc[:, i % self.period2, :])) + (1 - alphas) * (s + t)
            tnew = betas * (snew - s) + (1 - betas) * t

            Ico = Ic.clone()
            wco = wc.clone()

            Ico[:, i % self.period1, :] = gammas * (y[:, i, :] / (snew * wc[:, i % self.period2, :])) + (1 - gammas) * Ic[:, i % self.period1, :]
            wco[:, i % self.period2, :] = omegas * (y[:, i, :] / (snew * Ic[:, i % self.period1, :])) + (1 - omegas) * wc[:, i % self.period2, :]
            Ic = Ico
            wc = wco

            s = snew
            t = tnew

        return yhat, Ic, wc, s, t

    def forward(self, y, state=None):
        # We need 3 dimentions: (batch,series_len,feature)
        assert y.ndimension() == 3, "3-dim input is expected"
        assert y.size(1) >= 2 * self.period2, "Input series too short"

        if not state:
            if self.is_multiplicative:
                state = self.mult_init_params(y)
            else:
                state = self.add_init_params(y)
        Ic, wc, t, s = state

        phis = torch.sigmoid(self.phis)

        # Allocate space
        bs, n, f = y.size()
        e = torch.empty((bs, n, f))
        max_h = self.hs.max()

        if self.is_multiplicative:
            yhat, Ic, wc, s, t = self.mult_update(y, Ic, wc, s, t)
        else:
            yhat, Ic, wc, s, t = self.add_update(y, Ic, wc, s, t)

        # we need to shift seasonality as it doesn't match the last point when we finished updating it.
        # this happens when n%period!=0, so we have to roll by this amount, shifting last elements to be the
        # first ones
        Ic = torch.roll(Ic, -n % self.period1, dims=1)
        wc = torch.roll(wc, -n % self.period2, dims=1)

        e = y - yhat

        # do not compute fcast if we're not asked for it (h=0), we can do only yhat fit
        if max_h > 0:

            ca = s.view(bs, 1, f) + torch.arange(1, max_h + 1).float().view(1, -1, 1).repeat(bs, 1, f) * t.view(bs, 1, f)
            cb = Ic.repeat(1, max_h // self.period1 + 1, 1)[:, :max_h, :]
            cc = wc.repeat(1, max_h // self.period2 + 1, 1)[:, :max_h, :]

            if self.is_multiplicative:
                fcast = ca * cb * cc
            else:
                fcast = ca + cb + cc

            # TODO: additive
            if self.enable_ar:
                # trim too small values
                phis = torch.where(phis < 1e-7, torch.zeros(phis.size()), phis)
                # trim too big values
                phis = torch.where(phis > 0.9, torch.zeros(phis.size()), phis)

                # applied only if phi is bigger than 0, otherwise the result stays the same
                yhat = yhat + phis * torch.cat([torch.zeros((bs, 1, f)), e[:, :-1, :]], dim=1)
                fcast = fcast + phis ** torch.arange(1, max_h + 1).float().view(1, -1, 1).repeat(bs, 1, f) * e[:, n - 1, :].view(bs, 1, f)
                e = y - yhat
        else:
            fcast = torch.empty((bs, max_h, f))

        return make_obj(mean=fcast,
                        yhat=yhat,
                        state=(Ic, wc, t, s),
                        e=e)

    def get_loss(self, x, y, loss_fn, state=None):

        ret = self.forward(x, state=state)

        mse = torch.mean(ret.e[:, :, 0]**2)

        mseh = loss_fn(ret.mean[:, :, 1], y[:, :, 1])

        rloss = (mse + mseh)

        return ret.mean, rloss



import numpy as np

import torch

import pytest

from torches.modules.dshw_module import DSHWModule
from .utils.r_model import get_r_model


def test_r_vs_torch_noar(dataset_one, model_coeffs_noar):

    armethod = False

    r_reference_model = get_r_model(dataset_one.ts,
                                    dataset_one.period1,
                                    dataset_one.period2,
                                    dataset_one.h,
                                    model_coeffs_noar,
                                    armethod=armethod)

    torch_model = DSHWModule(input_dim=1,
                             period1=dataset_one.period1,
                             period2=dataset_one.period2,
                             hs=[dataset_one.h],
                             enable_ar=armethod,
                             alphas=[model_coeffs_noar['alpha']],
                             betas=[model_coeffs_noar['beta']],
                             gammas=[model_coeffs_noar['gamma']],
                             omegas=[model_coeffs_noar['omega']],
                             phis=[model_coeffs_noar['phi']])(dataset_one.torch_ts.view(1, -1, 1))

    assert np.allclose(torch_model.mean.detach().numpy().squeeze(), r_reference_model.mean)
    assert np.allclose(torch_model.yhat.detach().numpy().squeeze(), r_reference_model.fitted)


def test_r_vs_torch_ar(dataset_one, model_coeffs_ar):

    armethod = True

    r_reference_model = get_r_model(dataset_one.ts,
                                    dataset_one.period1,
                                    dataset_one.period2,
                                    dataset_one.h,
                                    model_coeffs_ar,
                                    armethod=armethod)

    torch_model = DSHWModule(input_dim=1,
                             period1=dataset_one.period1,
                             period2=dataset_one.period2,
                             hs=[dataset_one.h],
                             enable_ar=armethod,
                             alphas=[model_coeffs_ar['alpha']],
                             betas=[model_coeffs_ar['beta']],
                             gammas=[model_coeffs_ar['gamma']],
                             omegas=[model_coeffs_ar['omega']],
                             phis=[model_coeffs_ar['phi']])(dataset_one.torch_ts.view(1, -1, 1))

    assert np.allclose(torch_model.mean.detach().numpy().squeeze(), r_reference_model.mean)
    assert np.allclose(torch_model.yhat.detach().numpy().squeeze(), r_reference_model.fitted)


@pytest.mark.parametrize("is_multiplicative", [True, False])
def test_gradient_flow(is_multiplicative, dataset_one, model_coeffs_ar):

    torch_model = DSHWModule(input_dim=1,
                             period1=dataset_one.period1,
                             period2=dataset_one.period2,
                             hs=[dataset_one.h],
                             enable_ar=True,
                             is_multiplicative=is_multiplicative,
                             alphas=[model_coeffs_ar['alpha']],
                             betas=[model_coeffs_ar['beta']],
                             gammas=[model_coeffs_ar['gamma']],
                             omegas=[model_coeffs_ar['omega']],
                             phis=[model_coeffs_ar['phi']])(dataset_one.torch_ts.view(1, -1, 1))

    mse = torch.mean(torch_model.e[:, :, 0]**2)
    mse.backward()  # in case of in-place modification, this will raise RuntimeError


@pytest.mark.parametrize("enable_ar", [True, False])
@pytest.mark.parametrize("is_multiplicative", [True, False])
def test_two_series(enable_ar, is_multiplicative, dataset_2_feature, model_coeffs_ar):

    ref_1 = DSHWModule(input_dim=1,
                       period1=dataset_2_feature.period1,
                       period2=dataset_2_feature.period2,
                       hs=[dataset_2_feature.h],
                       enable_ar=enable_ar,
                       is_multiplicative=is_multiplicative,
                       alphas=[model_coeffs_ar['alpha']],
                       betas=[model_coeffs_ar['beta']],
                       gammas=[model_coeffs_ar['gamma']],
                       omegas=[model_coeffs_ar['omega']],
                       phis=[model_coeffs_ar['phi']])(dataset_2_feature.torch_ts[:, 0].view(1, -1, 1))

    ref_2 = DSHWModule(input_dim=1,
                       period1=dataset_2_feature.period1,
                       period2=dataset_2_feature.period2,
                       hs=[dataset_2_feature.h],
                       enable_ar=enable_ar,
                       is_multiplicative=is_multiplicative,
                       alphas=[model_coeffs_ar['alpha']],
                       betas=[model_coeffs_ar['beta']],
                       gammas=[model_coeffs_ar['gamma']],
                       omegas=[model_coeffs_ar['omega']],
                       phis=[model_coeffs_ar['phi']])(dataset_2_feature.torch_ts[:, 1].view(1, -1, 1))

    ref_all = DSHWModule(input_dim=2,
                         period1=dataset_2_feature.period1,
                         period2=dataset_2_feature.period2,
                         hs=[dataset_2_feature.h, dataset_2_feature.h],
                         enable_ar=enable_ar,
                         is_multiplicative=is_multiplicative,
                         alphas=[model_coeffs_ar['alpha'], model_coeffs_ar['alpha']],
                         betas=[model_coeffs_ar['beta'], model_coeffs_ar['beta']],
                         gammas=[model_coeffs_ar['gamma'], model_coeffs_ar['gamma']],
                         omegas=[model_coeffs_ar['omega'], model_coeffs_ar['omega']],
                         phis=[model_coeffs_ar['phi'], model_coeffs_ar['phi']]
                         )(dataset_2_feature.torch_ts.view(1, -1, 2))

    assert torch.allclose(torch.cat([ref_1.mean, ref_2.mean], dim=2), ref_all.mean)
    assert torch.allclose(torch.cat([ref_1.yhat, ref_2.yhat], dim=2), ref_all.yhat)
    assert torch.allclose(torch.cat([ref_1.e, ref_2.e], dim=2), ref_all.e)


@pytest.mark.parametrize("enable_ar", [True, False])
@pytest.mark.parametrize("is_multiplicative", [True, False])
def test_2_batches(enable_ar, is_multiplicative, dataset_2_batch, model_coeffs_ar):

    ref_1 = DSHWModule(input_dim=1,
                       period1=dataset_2_batch.period1,
                       period2=dataset_2_batch.period2,
                       hs=[dataset_2_batch.h],
                       enable_ar=enable_ar,
                       is_multiplicative=is_multiplicative,
                       alphas=[model_coeffs_ar['alpha']],
                       betas=[model_coeffs_ar['beta']],
                       gammas=[model_coeffs_ar['gamma']],
                       omegas=[model_coeffs_ar['omega']],
                       phis=[model_coeffs_ar['phi']])(dataset_2_batch.torch_ts[0, :].view(1, -1, 1))

    ref_2 = DSHWModule(input_dim=1,
                       period1=dataset_2_batch.period1,
                       period2=dataset_2_batch.period2,
                       hs=[dataset_2_batch.h],
                       enable_ar=enable_ar,
                       is_multiplicative=is_multiplicative,
                       alphas=[model_coeffs_ar['alpha']],
                       betas=[model_coeffs_ar['beta']],
                       gammas=[model_coeffs_ar['gamma']],
                       omegas=[model_coeffs_ar['omega']],
                       phis=[model_coeffs_ar['phi']])(dataset_2_batch.torch_ts[1, :].view(1, -1, 1))

    ref_all = DSHWModule(input_dim=1,
                         period1=dataset_2_batch.period1,
                         period2=dataset_2_batch.period2,
                         hs=[dataset_2_batch.h],
                         enable_ar=enable_ar,
                         is_multiplicative=is_multiplicative,
                         alphas=[model_coeffs_ar['alpha']],
                         betas=[model_coeffs_ar['beta']],
                         gammas=[model_coeffs_ar['gamma']],
                         omegas=[model_coeffs_ar['omega']],
                         phis=[model_coeffs_ar['phi']]
                         )(dataset_2_batch.torch_ts.view(2, -1, 1))

    assert torch.allclose(torch.cat([ref_1.mean, ref_2.mean], dim=2), ref_all.mean)
    assert torch.allclose(torch.cat([ref_1.yhat, ref_2.yhat], dim=2), ref_all.yhat)
    assert torch.allclose(torch.cat([ref_1.e, ref_2.e], dim=2), ref_all.e)


@pytest.mark.parametrize("enable_ar", [True, False])
@pytest.mark.parametrize("is_multiplicative", [True, False])
def test_two_batches_two_series(enable_ar, is_multiplicative, dataset_2_batch_2_feature, model_coeffs_ar):

    ref_1 = DSHWModule(input_dim=1,
                       period1=dataset_2_batch_2_feature.period1,
                       period2=dataset_2_batch_2_feature.period2,
                       hs=[dataset_2_batch_2_feature.h],
                       enable_ar=enable_ar,
                       is_multiplicative=is_multiplicative,
                       alphas=[model_coeffs_ar['alpha']],
                       betas=[model_coeffs_ar['beta']],
                       gammas=[model_coeffs_ar['gamma']],
                       omegas=[model_coeffs_ar['omega']],
                       phis=[model_coeffs_ar['phi']])(dataset_2_batch_2_feature.torch_ts[:, :, 0:1])

    ref_2 = DSHWModule(input_dim=1,
                       period1=dataset_2_batch_2_feature.period1,
                       period2=dataset_2_batch_2_feature.period2,
                       hs=[dataset_2_batch_2_feature.h],
                       enable_ar=enable_ar,
                       is_multiplicative=is_multiplicative,
                       alphas=[model_coeffs_ar['alpha']],
                       betas=[model_coeffs_ar['beta']],
                       gammas=[model_coeffs_ar['gamma']],
                       omegas=[model_coeffs_ar['omega']],
                       phis=[model_coeffs_ar['phi']])(dataset_2_batch_2_feature.torch_ts[:, :, 1:2])

    ref_all = DSHWModule(input_dim=2,
                         period1=dataset_2_batch_2_feature.period1,
                         period2=dataset_2_batch_2_feature.period2,
                         hs=[dataset_2_batch_2_feature.h, dataset_2_batch_2_feature.h],
                         enable_ar=enable_ar,
                         is_multiplicative=is_multiplicative,
                         alphas=[model_coeffs_ar['alpha'], model_coeffs_ar['alpha']],
                         betas=[model_coeffs_ar['beta'], model_coeffs_ar['beta']],
                         gammas=[model_coeffs_ar['gamma'], model_coeffs_ar['gamma']],
                         omegas=[model_coeffs_ar['omega'], model_coeffs_ar['omega']],
                         phis=[model_coeffs_ar['phi'], model_coeffs_ar['phi']]
                         )(dataset_2_batch_2_feature.torch_ts)

    assert torch.allclose(torch.cat([ref_1.mean, ref_2.mean], dim=2), ref_all.mean)
    assert torch.allclose(torch.cat([ref_1.yhat, ref_2.yhat], dim=2), ref_all.yhat)
    assert torch.allclose(torch.cat([ref_1.e, ref_2.e], dim=2), ref_all.e)



import torch


def pytorch_moving_average(a, n=2):
    assert a.ndimension() == 3, "3-dim input is expected"

    ret = torch.cumsum(a, dim=1)
    retc = ret.clone()
    retc[:, n:, :] = ret[:, n:, :] - ret[:, :-n, :]
    return retc[:, n - 1:, :] / n


# TODO: test this against R

# Calculating seasonal indexes
def mult_seasindex(y, p):

    assert y.ndimension() == 3, "3-dim input is expected"

    bs = y.size(0)
    n = y.size(1)
    f = y.size(2)

    n2 = 2 * p
    shorty = y[:, :n2, :]
    average = torch.empty((bs, n, f))
    simplema = pytorch_moving_average(shorty, p)

    if p % 2 == 0:  # Even order
        centeredma = pytorch_moving_average(simplema[:, 0:(n2 - p + 1), :], 2)
#         print(shorty.shape, average[p//2:p].shape, shorty[p//2:p].shape, centeredma[:p].shape)
        offset = p // 2
        averagec = average.clone()
        averagec[:, offset:offset + p, :] = shorty[:, offset:offset + p, :] / centeredma[:, :p, :]
        average = averagec
        ii = list(range(p, (p + p // 2))) + list(range((p // 2), p))
        si = average[:, ii, :]  # average[[p + (:(p / 2)), (p / 2):p]]
    else:
        offset = (p - 1) // 2
        average[:, offset:offset + p, :] = shorty[:, offset:offset + p, :] / simplema[:, :p, :]
        ii = list(range(p, (p + (p - 1) // 2))) + list(range((p // 2), p))
        si = average[:, ii, :]
        raise Exception("Not tested")
    return si


def add_seasindex(y, p):

    assert y.ndimension() == 3, "3-dim input is expected"

    bs = y.size(0)
    n = y.size(1)
    f = y.size(2)

    n2 = 2 * p
    shorty = y[:, :n2, :]
    average = torch.empty((bs, n, f))
    simplema = pytorch_moving_average(shorty, p)

    if p % 2 == 0:  # Even order
        centeredma = pytorch_moving_average(simplema[:, 0:(n2 - p + 1), :], 2)
#         print(shorty.shape, average[p//2:p].shape, shorty[p//2:p].shape, centeredma[:p].shape)
        offset = p // 2
        averagec = average.clone()
        averagec[:, offset:offset + p, :] = shorty[:, offset:offset + p, :] - centeredma[:, :p, :]
        average = averagec
        ii = list(range(p, (p + p // 2))) + list(range((p // 2), p))
        si = average[:, ii, :]  # average[[p + (:(p / 2)), (p / 2):p]]
    else:
        offset = (p - 1) // 2
        average[:, offset:offset + p, :] = shorty[:, offset:offset + p, :] - simplema[:, :p, :]
        ii = list(range(p, (p + (p - 1) // 2))) + list(range((p // 2), p))
        si = average[:, ii, :]
        raise Exception("Not tested")
    return si



import pytest

import rpy2.robjects as ro

from .utils.fixtures import DataSetOneFixture, DataSet2FeatureFixture, DataSet2BatchFixture, DataSet2Batch2FeatureFixture


@pytest.fixture(scope="module")
def dataset_one():
    return DataSetOneFixture()


@pytest.fixture(scope="module")
def dataset_2_feature():
    return DataSet2FeatureFixture()


@pytest.fixture(scope="module")
def dataset_2_batch():
    return DataSet2BatchFixture()


@pytest.fixture(scope="module")
def dataset_2_batch_2_feature():
    return DataSet2Batch2FeatureFixture()


@pytest.fixture(scope="module")
def model_coeffs_noar():
    return {
        'alpha': 0.1648906365819534,
        'beta': 0.08265189925077121,
        'gamma': 2.398595300137416e-07,
        'omega': 0.2309265700845013,
        'phi': 0.0,
        'lambda': ro.r("NULL")
    }


@pytest.fixture(scope="module")
def model_coeffs_ar():
    return {
        'alpha': 0.1648906365819534,
        'beta': 0.08265189925077121,
        'gamma': 2.398595300137416e-07,
        'omega': 0.2309265700845013,
        'phi': 0.1448967,
        'lambda': ro.r("NULL")
    }



import torch


def pytorch_diff(inp):
    assert inp.ndimension() == 3, "3-dim input is expected"
    return inp[:, 1:, :] - inp[:, :-1, :]


def default_param(input_dim, param_value, default_value, requires_grad=True):
    if param_value is None:
        return siginv(torch.tensor([default_value] * input_dim, requires_grad=requires_grad))
    else:
        assert len(param_value) == input_dim
        return siginv(torch.tensor([param_value], requires_grad=requires_grad))


def siginv(x):
    return torch.log(x / (1 - x))



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="torch-es",
    version="0.0.1",
    author="Vladimir Shulyak",
    author_email="vladimir@shulyak.net",
    description="Double Seasonal Exponential Smoothing using PyTorch + ES-RNN capabilities on top",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/vshulyak/torch-es",
    packages=setuptools.find_packages(),
    install_requires=[
        'torch>=1.0.0',
        'numpy>=1.16.0'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
